#!/usr/bin/env python
# coding=utf-8
import sys, getopt, numpy, math, itertools
# from itertools import combinations

def union_key(str1, str2):
	if str1 > str2:
		return str2 + "|" + str1
	else:
		return str1 + "|" +str2

# calculate conditional mutual information
def calculate_cmi(list1, list2, conditional_lists):
	# get a copy of conditional_lists
	l1 = conditional_lists[:]
	l2 = conditional_lists[:]
	l3 = conditional_lists[:]
	l1.append(list1)
	l2.append(list2)
	l3.append(list1)
	l3.append(list2)
	cmi = numpy.linalg.det(numpy.cov(l1)) * numpy.linalg.det(numpy.cov(l2))
	if len(conditional_lists) == 1:
		cmi = cmi / numpy.cov(conditional_lists)
	else:
		cmi = cmi / numpy.linalg.det(numpy.cov(conditional_lists))

	cmi = cmi / numpy.linalg.det(numpy.cov(l3))
	cmi = 0.5*math.log(cmi)
	return cmi

# calculate mutual information
def calculate_mi(list1, list2):
	if len(list1) != len(list2):
		exit()
	# co_occurance filter 80%
	co_occurance = 0.0
	for x in range(0,len(list1)):
		if list1[x] != 0  and list2[x] != 0:
			co_occurance = co_occurance+1

	if co_occurance / len(list1) < 0.8:
		return 0

	mi = numpy.var(list1) * numpy.var(list2)
	try:
		mi = mi / numpy.linalg.det(numpy.cov(list1, list2))
	except ZeroDivisionError:
		return 0
	
	mi = 0.5 * math.log(mi)
	return mi
	

def list_intersection(list1, list2):
	intersection = []
	for x in list1:
		try:
			list2.index(x)
		except Exception as e:
			pass
		else:
			intersection.append(x)
	return intersection





def pca_cmi(argv):
	input_file = ''
	output_file = ''
	threshold = 0
	try:
		opts, args = getopt.getopt(argv, "hi:o:t:", ["ifile=","ofile=","threshold="])
		# print opts
	except getopt.GetoptError:
		print "matrix_trans -i <input_file> -t <threshold> -o <output_file>"
		sys.exit()
	for opt, arg in opts:
		if opt == '-h':
			print "matrix_trans -i <input_file> -o <output_file>"
			sys.exit()
		elif opt in ("-i", "--ifile"):
			input_file = arg
		elif opt in ("-o", "--ofile"):
			output_file = arg
			# print arg
		elif opt in ("-t", "--threshold"):
			threshold = abs(float(arg))
			# print arg
	if len(output_file) == 0:
		output_file = input_file + "_pca_cmi.out"
	# print output_file
	if len(input_file) == 0:
		print "matrix_trans -i <input_file> -o <output_file>"
		sys.exit(2)

	ifile = open(input_file,"r")
	ofile = open(output_file,"w")
	# print threshold
	# skip the first line
	ifile.next();

	abundance_dict = {}
	for line in ifile:
		line = line.rstrip("\n");
		line = line.rstrip("\t");
		row = line.split("\t")
		abundance_dict[row[0]] = row[1:]
		for x in range(0, len(abundance_dict[row[0]]) ):
			abundance_dict[row[0]][x] = float(abundance_dict[row[0]][x])
			# pass


	# print len(abundance_dict)
	# a = 0
	# print abundance_dict.index("s")
	abundance_dict_keys = abundance_dict.keys()
	species_count = len(abundance_dict)
	corrlation_dict = {}
	print "total species count:",species_count
	print "total corrlation count:",species_count * (species_count - 1) / 2
	# print abundance_dict["Candidatus_Portiera;"]
	count = 1
	for i in range(0,species_count):
		for j in range(i+1, species_count):
			key1 = abundance_dict_keys[i]
			key2 = abundance_dict_keys[j]
			mi = calculate_mi(abundance_dict[key1],abundance_dict[key2])
			# print mi
			sys.stdout.write('calulating: %s\r' % str(count))
			sys.stdout.flush()
			count = count + 1
			if abs(mi) >= threshold and mi != 0:
				key = union_key(key1,key2)
				corrlation_dict[key] = mi

	print "zreo-order corrlation count:", len(corrlation_dict)
	# corrlation_dict["a|b"] = 0.1
	# neighbors_dict[a] = ["b","c"]
	neighbors_dict = {}
	corrlation_dict_keys = corrlation_dict.keys()
	for key in corrlation_dict_keys:
		keys = key.split("|")
		if neighbors_dict.has_key(keys[0]):
			neighbors_dict[keys[0]].append(keys[1])
		else:
			neighbors_dict[keys[0]] = []
		if neighbors_dict.has_key(keys[1]):
			neighbors_dict[keys[1]].append(keys[0])
		else:
			neighbors_dict[keys[1]] = []

	# print neighbors_dict
	# print len(neighbors_dict)
	# 可以在此处释放内存

	order = 0
	edge_reducing = 0
	# PCA
	while order < 10:
		order = order + 1
		print "order:",order
		print "before:",len(corrlation_dict_keys)
		for key in corrlation_dict_keys:
			keys = key.split("|")
			sp1 = keys[0]
			sp2 = keys[1]
			sp1_neighbors = neighbors_dict[sp1]
			sp2_neighbors = neighbors_dict[sp2]
			neighbors_intersection = list_intersection(sp1_neighbors, sp2_neighbors)
			if len(neighbors_intersection) >= order:
				conditions_species = itertools.combinations(neighbors_intersection,order)
				conditions_species = list(conditions_species)
				# print conditions
				for condition in conditions_species:
					condition_list = []
					for x in condition:
						condition_list.append(abundance_dict[x])

					cmi = calculate_cmi(abundance_dict[sp1],abundance_dict[sp2],condition_list)
					# print cmi
					if abs(cmi) == 0:
						
						k = union_key(sp1,sp2)
						if corrlation_dict.has_key(k):
							corrlation_dict.pop(k)
							edge_reducing = edge_reducing + 1

		print "edge reduced:",edge_reducing
		print "after: ",len(corrlation_dict)
		corrlation_dict_keys = corrlation_dict.keys()
		if edge_reducing == 0:
			break

		edge_reducing = 0
		print
		# print sp1,sp2
		# print sp1_neighbors
		# print sp2_neighbors
		
		# print neighbors_intersection
		
		# pass
	# print corrlation_dict
	for key in corrlation_dict_keys:
		keys = key.split("|")
		sp1 = keys[0]
		sp2 = keys[1]
		out_str = sp1 + "\t" + sp2 + "\t" + str(corrlation_dict[union_key(sp1,sp2)])
		ofile.write(out_str)
		ofile.write("\n")
		# ofile.write
		# print out_str

	# print a
	# print len(abundance_dict)*len(abundance_dict)
	ifile.close()
	ofile.close()
	

if __name__ == "__main__":
	pca_cmi(sys.argv[1:])
	# l1 = [1,2]
	# l2 = 'asd'
	# l3 = [2,3,6,9]
	# l4 = [1,4,2,6]
	# # print mi(l1,l2)
	# # cmi(l1,l2,[l4])
	# union_key("aaa","bbb")
	# print combination(l1,1)
	# iter = itertools.combinations(l1,1)
	# l = list(iter)
	# print l
